from typing import List, Dict, Any, Union

from langchain.agents import AgentOutputParser
from langchain.schema import AgentAction

from yid_langchain_extensions.output_parser.thoughts_json_parser import Thought, ThoughtsJSONParser


class ActionWithThoughts(AgentAction):
    tool_input: Union[str, list, dict]
    all_thoughts: Dict[str, Any]


class ActionParser(ThoughtsJSONParser, AgentOutputParser):
    @classmethod
    def get_action_thoughts(cls, type_of_action_input: str) -> List[Thought]:
        return [
            Thought(name="action", description="The action to take. Must be one of [{{tool_names}}]"),
            Thought(name="action_input", description="The input to the action", type=type_of_action_input)
        ]

    @classmethod
    def from_extra_thoughts(cls, pre_thoughts: List[Thought], after_thoughts: List[Thought],
                            type_of_action_input: str = "string"):
        thoughts = pre_thoughts + cls.get_action_thoughts(type_of_action_input) + after_thoughts
        return cls(thoughts=thoughts)

    def parse(self, text: str) -> ActionWithThoughts:
        response = super().parse(text)
        fixed_text = self.fix_json_md_snippet(text)
        return ActionWithThoughts(
            tool=response["action"], tool_input=response["action_input"], log=fixed_text, all_thoughts=response)



from typing import List, Dict
from unittest import TestCase

from langchain_community.llms.fake import FakeListLLM
from langchain_core.language_models import LLM
from langchain_core.messages import BaseMessage
from langchain_core.runnables import RunnableSerializable
from pydantic import BaseModel

from yid_langchain_extensions.utils import pydantic_v1_port


class Pydantic2ClassWithoutPort(BaseModel):
    llm: LLM


class Pydantic2ClassWithPort(BaseModel):
    llm: pydantic_v1_port(LLM)


class Pydantic2ClassWithPortNested(BaseModel):
    llm_list: pydantic_v1_port(List[LLM])
    llm_dict: pydantic_v1_port(Dict[str, LLM])
    complex_field: pydantic_v1_port(RunnableSerializable[List[BaseMessage], str])


class TestPydanticV1Port(TestCase):
    def setUp(self):
        self.llm = FakeListLLM(responses=[])

    def test_v1_port(self):
        Pydantic2ClassWithPort(llm=self.llm)

    def test_port_required(self):
        with self.assertRaises(TypeError):
            Pydantic2ClassWithoutPort(llm=self.llm)

    def test_v1_port_nested(self):
        Pydantic2ClassWithPortNested(
            llm_list=[self.llm],
            llm_dict={"llm": self.llm},
            complex_field=self.llm
        )



from typing import List

from langchain import PromptTemplate
from langchain.tools import BaseTool

TOOLS_PROMPT = """AVAILABLE ACTIONS:
------
Now you have access to following actions in order to solve the task:
{tools}"""


def format_tools(tools: List[BaseTool]) -> str:
    tools_string = "\n".join(
        [f"> {tool.name}: {tool.description}" for tool in tools]
    )
    return PromptTemplate.from_template(TOOLS_PROMPT).format_prompt(
        tools=tools_string
    ).to_string()


def format_tool_names(tools: List[BaseTool]) -> str:
    return ", ".join([tool.name for tool in tools])


class FinalAnswerTool(BaseTool):
    name: str = "final_answer"
    description: str = "Use this if you want to respond directly to the human."
    return_direct: bool = True

    def _run(self, action_input: str) -> str:
        return action_input

    async def _arun(self, action_input: str) -> str:
        return action_input



import unittest

from langchain import PromptTemplate
from langchain.llms import FakeListLLM
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate, MessagesPlaceholder
from langchain.schema import SystemMessage
from langchain.tools import Tool

from yid_langchain_extensions.agent.simple_agent import SimpleAgent
from yid_langchain_extensions.output_parser.action_parser import ActionParser
from yid_langchain_extensions.tools.utils import format_tools, format_tool_names, FinalAnswerTool


class TestSimpleAgent(unittest.TestCase):
    def test_run(self):
        answers = [
            '```json\n{\n\t"action": "check_weather",\n\t"action_input": "Moscow"\n',
            '```json\n{\n\t"action": "final_answer",\n\t"action_input": "In Moscow rainy with a temperature of 10°C."\n'
        ]
        llm = FakeListLLM(responses=answers)
        weather_tool = Tool(
            name="check_weather", description="Use it to check weather at some location",
            func=lambda x: "Rain, 10 C"
        )
        final_answer_tool = FinalAnswerTool()
        tools = [weather_tool, final_answer_tool]
        output_parser = ActionParser.from_extra_thoughts([], [])
        template = ChatPromptTemplate.from_messages(
            messages=[
                HumanMessagePromptTemplate.from_template("{{input}}", "jinja2"),
                MessagesPlaceholder(variable_name="agent_scratchpad"),
                SystemMessage(content=format_tools(tools)),
                SystemMessage(content=PromptTemplate.from_template(
                    output_parser.get_format_instructions(), template_format="jinja2").format(
                    tool_names=format_tool_names(tools)
                ))
            ]
        )
        agent_executor = SimpleAgent.from_llm_and_prompt(
            llm=llm, prompt=template, output_parser=output_parser, stop_sequences=output_parser.stop_sequences
        ).get_executor(tools=tools)
        answer = agent_executor.run(input="What is the weather in Moscow?")
        self.assertEqual(answer, "In Moscow rainy with a temperature of 10°C.")



import unittest

from pydantic.v1 import ValidationError

from yid_langchain_extensions.output_parser.action_parser import ActionParser
from yid_langchain_extensions.output_parser.class_parser import Class, get_classes_description, get_classes_summary, \
    ClassParser
from yid_langchain_extensions.output_parser.thoughts_json_parser import ThoughtsJSONParser, Thought


class TestThoughtsJSONParser(unittest.TestCase):
    def test_parse(self):
        parser = ThoughtsJSONParser(
            thoughts=[
                Thought(name="thought1", description="Thought 1"),
                Thought(name="thought2", type="int", description="Thought 2")
            ]
        )
        string_to_parse = """```json
{
    "thought1": "thought 1",
    "thought2": 2
"""
        parsed = parser.parse(string_to_parse)
        self.assertEqual(parsed, {
            "thought1": "thought 1",
            "thought2": 2
        })

    def test_parse_without_json(self):
        parser = ThoughtsJSONParser(
            thoughts=[
                Thought(name="thought1", description="Thought 1")
            ]
        )
        string_to_parse = """{
    "thought1": "thought 1"
}
"""
        parsed = parser.parse(string_to_parse)
        self.assertEqual(parsed, {
            "thought1": "thought 1"
        })

    def test_unpaired_brackets(self):
        parser = ThoughtsJSONParser(
            thoughts=[
                Thought(name="thought1", description="Thought 1")
            ]
        )
        string_to_parse = """```json
{
    "thought1": "thought 1"
}
"""
        parsed = parser.parse(string_to_parse)
        self.assertEqual(parsed, {
            "thought1": "thought 1"
        })

    def test_parse_multi_line_value(self):
        parser = ThoughtsJSONParser(
            thoughts=[
                Thought(name="thought1", description="Thought 1")
            ]
        )
        string_to_parse = """{
        "thought1": "thought
1"
    }
    """
        parsed = parser.parse(string_to_parse)
        self.assertEqual(parsed, {
            "thought1": "thought\n1"
        })

    def test_format_thoughts(self):
        thoughts = [
            Thought(name="thought1", description="Thought 1"),
            Thought(name="thought2", type="int", description="Thought 2")
        ]
        thoughts_string = ThoughtsJSONParser(
            thoughts=thoughts
        ).format_thoughts()
        self.assertEqual(thoughts_string, '"thought1": string [Thought 1]\n\t"thought2": int [Thought 2]')

    def test_format_instructions(self):
        thoughts = [
            Thought(name="thought1", description="Thought 1"),
            Thought(name="thought2", type="int", description="Thought 2")
        ]
        format_instructions = ThoughtsJSONParser(
            thoughts=thoughts
        ).get_format_instructions()
        self.assertTrue('"thought1": string [Thought 1]\n\t"thought2": int [Thought 2]' in format_instructions)

    def test_complex_format_instructions(self):
        thoughts = [
            Thought(name="tool", description="One of [{{tools}}]"),
        ]
        format_instructions = ThoughtsJSONParser(
            thoughts=thoughts
        ).get_format_instructions(tools="tool1, tool2")
        self.assertTrue(format_instructions, '"tool": string [One of [tool1, tool2]]' in format_instructions)

    def test_raise_without_thoughts(self):
        with self.assertRaises(ValidationError):
            ThoughtsJSONParser(thoughts=[])

    def test_code_block(self):
        parser = ThoughtsJSONParser(
            thoughts=[
                Thought(name="action", description="action"),
                Thought(name="action_input", description="action_input")
            ]
        )
        string_to_parse = '''```json
        {
            "action": "Final Answer",
                "action_input": "Example:\\n```python\\nimport cv2\\n```\\n"
                '''
        parsed = parser.parse(string_to_parse)
        self.assertEqual(parsed["action_input"], "Example:\n```python\nimport cv2\n```\n")


class TestClassParser(unittest.TestCase):
    def setUp(self) -> None:
        self.classes = [
            Class(name="class1", description="description1"),
            Class(name="class2", description="description2"),
        ]

    def test_get_classes_description(self):
        self.assertEqual(get_classes_description(self.classes),
                         "> 0: class1. description1;\n> 1: class2. description2;\n")

    def test_get_classes_summary(self):
        self.assertEqual(get_classes_summary(self.classes), "0 (class1); 1 (class2)")

    def test_class_parser(self):
        class_parser = ClassParser.from_extra_thoughts(extra_thoughts=[
            Thought(name="thought1", description="Thought 1")
        ])
        string_to_parse = """```json
{
    "thought1": "thought 1",
    "class_index": 0
"""
        self.assertEqual(class_parser.parse(string_to_parse), 0)

    def test_extra_thoughts(self):
        extra_thoughts = [
            Thought(name="thought1", description="Thought 1"),
            Thought(name="thought2", type="int", description="Thought 2")
        ]
        class_parser = ClassParser.from_extra_thoughts(extra_thoughts=extra_thoughts)
        self.assertEqual(class_parser.thoughts[:2], extra_thoughts)
        self.assertEqual(class_parser.thoughts[2].name, "class_index")


class TestActionParser(unittest.TestCase):
    def test_action_parser(self):
        action_parser = ActionParser.from_extra_thoughts(pre_thoughts=[
            Thought(name="thought1", description="Thought 1")
        ], after_thoughts=[
            Thought(name="thought2", description="Thought 2")
        ])
        string_to_parse = """```json
{
    "thought1": "thought 1",
    "action": "action 1",
    "action_input": "action input",
    "thought2": "thought 2"
"""
        action = action_parser.parse(string_to_parse)
        self.assertEqual(action.tool, "action 1")
        self.assertEqual(action.tool_input, "action input")
        self.assertEqual(action.log, string_to_parse + "}\n```")
        self.assertEqual(action.all_thoughts["thought1"], "thought 1")
        self.assertEqual(action.all_thoughts["thought2"], "thought 2")

    def test_extra_thoughts(self):
        pre_thoughts = [
            Thought(name="thought1", description="Thought 1"),
            Thought(name="thought2", type="int", description="Thought 2")
        ]
        after_thoughts = [
            Thought(name="thought3", description="Thought 5"),
            Thought(name="thought4", type="int", description="Thought 6")
        ]
        action_parser = ActionParser.from_extra_thoughts(pre_thoughts=pre_thoughts, after_thoughts=after_thoughts)
        self.assertEqual(action_parser.thoughts[:2], pre_thoughts)
        self.assertEqual(action_parser.thoughts[2].name, "action")
        self.assertEqual(action_parser.thoughts[3].name, "action_input")
        self.assertEqual(action_parser.thoughts[4:], after_thoughts)

    def test_list_action_input(self):
        action_parser = ActionParser.from_extra_thoughts(pre_thoughts=[], after_thoughts=[])
        string_to_parse = """```json
{
    "action": "action 1",
    "action_input": ["action input 1", "action input 2"]
"""
        action = action_parser.parse(string_to_parse)
        self.assertEqual(action.tool, "action 1")
        self.assertEqual(action.tool_input,  ["action input 1", "action input 2"])
        self.assertEqual(action.log, string_to_parse + "}\n```")

    def test_list_action_dict(self):
        action_parser = ActionParser.from_extra_thoughts(pre_thoughts=[], after_thoughts=[])
        string_to_parse = """```json
{
    "action": "action 1",
    "action_input": {"parameter_name": "parameter_value"}
"""
        action = action_parser.parse(string_to_parse)
        self.assertEqual(action.tool, "action 1")
        self.assertEqual(action.tool_input,  {"parameter_name": "parameter_value"})
        self.assertEqual(action.log, string_to_parse + "}\n```")



from typing import Annotated, Type, Any

from pydantic import ValidationError, PlainValidator
from pydantic.v1 import BaseModel as BaseModelV1, parse_obj_as


def validated(value: BaseModelV1, target_class: Type[Any]) -> BaseModelV1:
    try:
        return parse_obj_as(target_class, value)  # noqa
    except ValidationError as e:
        raise ValidationError(f"Input of invalid type: {e}")


def pydantic_v1_port(cls):
    return Annotated[cls, PlainValidator(lambda val: validated(val, cls))]
